package org.zeroturnaround.javarebel.integration.servlet;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;
import org.zeroturnaround.javarebel.integration.servlet.util.IoUtil;
import org.zeroturnaround.javarebel.integration.servlet.util.MiscUtil;
import org.zeroturnaround.javarebel.integration.servlet.util.ServletUtil;

/**
 * Servlet that provides resources using the <code>ServletContext</code>.
 * <p>
 * The purpose of this servlet is to provide resources accessed through HTTP
 * by using the <code>ServletContext</code> interface instead of any
 * container specific mechanisms. This enables JavaRebel to handle all
 * resources accessed through HTTP independent of the specific container.
 * </p>
 * <p>
 * This servlet should be mapped as the default servlet:
 * <pre>
 * &lt;servlet&gt;
 *   &lt;servlet-name&gt;defaultServlet&lt;/servlet-name&gt;
 *   &lt;servlet-class&gt;org.zeroturnaround.javarebel.integration.servlet.JavaRebelResourceServlet&lt;/servlet-class&gt;
 * &lt;/servlet&gt;
 *
 * &lt;servlet-mapping&gt;
 *   &lt;servlet-name&gt;defaultServlet&lt;/servlet-name&gt;
 *   &lt;url-pattern&gt;/&lt;/url-pattern&gt;
 * &lt;/servlet-mapping&gt;
 * </pre>
 * 
 * @author Rein Raudjärv
 * 
 * @see ServletContext
 */
public class JavaRebelResourceServlet extends HttpServlet {

  private static final long serialVersionUID = 1L;

  private static final Logger log = LoggerFactory.getInstance();

  // TODO parse web.xml
  private static final String[] welcomeFiles = { "welcome.html", "index.html", "index.htm", "index.jsp" };

  public void init() {
    log.log("JavaRebelResourceServlet initialized");
  }

  public void destroy() {
    log.log("JavaRebelResourceServlet destroyed");
  }

  protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
    try {
      process(req, resp);
    }
    catch (Exception e) {
      throw new ServletException(e.getMessage(), e);
    }
  }

  protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
    try {
      process(req, resp);
    }
    catch (Exception e) {
      throw new ServletException(e.getMessage(), e);
    }
  }

  private void process(HttpServletRequest req, HttpServletResponse resp) throws Exception {
    final String uri = ServletUtil.getUriFromRequest(req);
    final ServletContext servletContext = getServletContext();

    if (!ServletUtil.exists(servletContext, uri)) {
      // Resource not found
      resp.sendError(HttpServletResponse.SC_NOT_FOUND);
      return;
    }

    InputStream in = servletContext.getResourceAsStream(uri);
    if (in != null) {
      // File found
      sendData(req, resp, in);
      return;
    }
    // Directory found

    if (!uri.endsWith("/")) {
      // Trying to add '/' to end
      redirect(req, resp, uri + "/");
      return;
    }
    // Path ends with '/' character

    // Redirect to a welcome file
    for (int i = 0; i < welcomeFiles.length; i++) {
      String filename = welcomeFiles[i];
      String redirectPath = uri + filename;
      if (ServletUtil.exists(servletContext, redirectPath)) {
        // Found an existing welcome file
        redirect(req, resp, redirectPath);
        return;
      }
    }
    // No welcome files found

    // Show directory listing
    List paths = MiscUtil.removePrefix(uri, servletContext.getResourcePaths(uri));
    sendDirectory(req, resp, uri, paths);
    return;
  }

  /**
   * Redirect to a new URI.
   */
  private static void redirect(HttpServletRequest req, HttpServletResponse resp, String uri) throws IOException {
    String contextPath = req.getContextPath();
    resp.setContentLength(0);
    resp.sendRedirect(contextPath + uri);
  }

  /**
   * Send contents of an {@link InputStream}.
   */
  private static void sendData(HttpServletRequest req, HttpServletResponse resp, InputStream in) throws Exception {
    // Use available() method of ByteArrayInputStream
    if (in instanceof ByteArrayInputStream) {
      resp.setContentLength(in.available());
    }

    OutputStream out = resp.getOutputStream();
    try {
      IoUtil.copy(in, out);
    }
    finally {
      in.close();
    }
  }

  /**
   * Write out a directory listing.
   * 
   * @param path requested URI.
   * @param subPaths entry names in the requested directory.
   */
  private static void sendDirectory(HttpServletRequest req, HttpServletResponse resp, String uri, Collection subPaths) throws IOException {
    String contextPath = req.getContextPath();
    String path = contextPath + uri;
    
    resp.setContentType("text/html; charset=utf-8");
    
    PrintWriter out = resp.getWriter();
    
    String title = "Directory: " + path;
    String parentHref = ServletUtil.encodeUrl(ServletUtil.getParentDirectory(path));

    // Header
    out.print("<html><head>");
    
    // Content Type
    out.print("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />");
    
    // Title
    out.print("<title>");
    out.print(title);
    out.print("</title></head>");
    
    out.print("<body><h1>");
    out.print(title);
    out.print("</h1><table border=\"0\">");
    
    // Parent directory
    out.print("<tr><td><a href=\"");
    out.print(parentHref);
    out.print("\">Parent Directory</a></td></tr>");

    Collection paths = new TreeSet(subPaths);
    for (Iterator it = paths.iterator(); it.hasNext();) {
      String subPath = (String) it.next();

      // Skip hidden files
      if (subPath.startsWith("."))
        continue;
      
      out.print("<tr><td><a href=\"");
      out.print(ServletUtil.encodeUrl(path + subPath));

      out.print("\">");
      out.print(subPath);
      out.print("</td></tr>");
    }

    // Footer
    out.print("</table></body></html>");
  }

}
